# Copyright (C) 2016-2016 Nicolas Sebrecht & contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

"""

The virtual imaplib2 takes care to import the correct imaplib2 library. Any
internal use of imaplib2 everywhere else in offlineimap must be done through
this virtual_imaplib2 or we might go into troubles.

"""

DESC = None

_SUPPORTED_RELEASE = 2
_SUPPORTED_REVISION = 57

try:
    # Try any imaplib2 in PYTHONPATH first. This allows both maintainers of
    # distributions and developers to not work with the bundled imaplib2.
    from imaplib2 import *
    import imaplib2 as imaplib

    if (int(imaplib.__release__) < _SUPPORTED_RELEASE or
            int(imaplib.__revision__) < _SUPPORTED_REVISION):
        raise ImportError("The provided imaplib2 version '%s' is not supported"%
            imaplib.__version__)
    DESC = "system"
except (ImportError, NameError) as e:
    try:
        from offlineimap.bundled_imaplib2 import *
        import offlineimap.bundled_imaplib2 as imaplib

        DESC = "bundled"
    except:
        print("Error while trying to import system imaplib2: %s"% e)
        raise

# Upstream won't expose those literals to avoid erasing them with "import *" in
# case they exist.
__version__ = imaplib.__version__
__release__ = imaplib.__release__
__revision__ = imaplib.__revision__



# Copyright 2013 Eygene A. Ryabinkin
# Functions to perform stack tracing (for multithreaded programs
# as well as for single-threaded ones).

import sys
import threading
import traceback


def dump(out):
	""" Dumps current stack trace into I/O object 'out' """
	id2name = {}
	for th in threading.enumerate():
		id2name[th.ident] = th.name
	n = 0
	for i, stack in sys._current_frames().items():
		out.write ("\n# Thread #%d (id=%d), %s\n" % \
		  (n, i, id2name[i]))
		n = n + 1
		for f, lno, name, line in traceback.extract_stack (stack):
			out.write ('File: "%s", line %d, in %s' % \
			  (f, lno, name))
			if line:
				out.write (" %s" % (line.strip()))
			out.write ("\n")



# Maildir repository support
# Copyright (C) 2002-2015 John Goerzen & contributors
# <jgoerzen@complete.org>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

from offlineimap.repository.Maildir import MaildirRepository
from offlineimap.folder.GmailMaildir import GmailMaildirFolder

class GmailMaildirRepository(MaildirRepository):
    def __init__(self, reposname, account):
        """Initialize a MaildirRepository object.  Takes a path name
        to the directory holding all the Maildir directories."""

        super(GmailMaildirRepository, self).__init__(reposname, account)

    def getfoldertype(self):
        return GmailMaildirFolder



# Base folder support
# Copyright (C) 2002-2016 John Goerzen & contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import os.path
import re
import time
from sys import exc_info

from offlineimap import threadutil
from offlineimap.ui import getglobalui
from offlineimap.error import OfflineImapError
import offlineimap.accounts


class BaseFolder(object):
    __hash__ = None

    def __init__(self, name, repository):
        """
        :param name: Path & name of folder minus root or reference
        :param repository: Repository() in which the folder is.
        """

        self.ui = getglobalui()
        self.messagelist = {}
        # Save original name for folderfilter operations.
        self.ffilter_name = name
        # Top level dir name is always ''.
        self.root = None
        self.name = name if not name == self.getsep() else ''
        self.newmail_hook = None
        # Only set the newmail_hook if the IMAP folder is named 'INBOX'.
        if self.name == 'INBOX':
            self.newmail_hook = repository.newmail_hook
        self.have_newmail = False
        self.copy_ignoreUIDs = None # List of UIDs to ignore.
        self.repository = repository
        self.visiblename = repository.nametrans(name)
        # In case the visiblename becomes '.' or '/' (top-level) we use
        # '' as that is the name that e.g. the Maildir scanning will
        # return for the top-level dir.
        if self.visiblename == self.getsep():
            self.visiblename = ''

        self.repoconfname = "Repository " + repository.name

        self.config = repository.getconfig()

        # Do we need to use mail timestamp for filename prefix?
        filename_use_mail_timestamp_global = self.config.getdefaultboolean(
            "general", "filename_use_mail_timestamp", False)
        self._filename_use_mail_timestamp = self.config.getdefaultboolean(
            self.repoconfname,
            "filename_use_mail_timestamp",
            filename_use_mail_timestamp_global)

        self._sync_deletes = self.config.getdefaultboolean(
            self.repoconfname, "sync_deletes", True)
        self._dofsync = self.config.getdefaultboolean("general", "fsync", True)

        # Determine if we're running static or dynamic folder filtering
        # and check filtering status.
        self._dynamic_folderfilter = self.config.getdefaultboolean(
            self.repoconfname, "dynamic_folderfilter", False)
        self._sync_this = repository.should_sync_folder(self.ffilter_name)
        if self._dynamic_folderfilter:
            self.ui.debug('', "Running dynamic folder filtering on '%s'[%s]"%
                (self.ffilter_name, repository))
        elif not self._sync_this:
            self.ui.debug('', "Filtering out '%s'[%s] due to folderfilter"%
                (self.ffilter_name, repository))

        # Passes for syncmessagesto.
        self.syncmessagesto_passes = [
            self.__syncmessagesto_copy,
            self.__syncmessagesto_delete,
            self.__syncmessagesto_flags,
        ]

    def getname(self):
        """Returns name"""
        return self.name

    def __str__(self):
        # FIMXE: remove calls of this. We have getname().
        return self.name

    def __unicode__(self):
        # NOTE(sheeprine): Implicit call to this by UIBase deletingflags() which
        # fails if the str is utf-8
        return self.name.decode('utf-8')

    def __enter__(self):
        """Starts a transaction. This will postpone (guaranteed) saving to disk
        of all messages saved inside this transaction until its committed."""
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Commits a transaction, all messages saved inside this transaction
        will only now be persisted to disk."""
        pass

    @property
    def accountname(self):
        """Account name as string"""

        return self.repository.accountname

    @property
    def sync_this(self):
        """Should this folder be synced or is it e.g. filtered out?"""

        if not self._dynamic_folderfilter:
            return self._sync_this
        else:
            return self.repository.should_sync_folder(self.ffilter_name)

    def dofsync(self):
        return self._dofsync

    def suggeststhreads(self):
        """Returns True if this folder suggests using threads for actions.

        Only IMAP returns True. This method must honor any CLI or configuration
        option."""

        return False

    def waitforthread(self):
        """Implements method that waits for thread to be usable.
        Should be implemented only for folders that suggest threads."""
        raise NotImplementedError

    def quickchanged(self, statusfolder):
        """ Runs quick check for folder changes and returns changed
        status: True -- changed, False -- not changed.

        :param statusfolder: keeps track of the last known folder state.
        """

        return True

    def getinstancelimitnamespace(self):
        """For threading folders, returns the instancelimitname for
        InstanceLimitedThreads."""

        raise NotImplementedError

    def storesmessages(self):
        """Should be true for any backend that actually saves message bodies.
        (Almost all of them).  False for the LocalStatus backend.  Saves
        us from having to slurp up messages just for localstatus purposes."""

        return 1

    def getvisiblename(self):
        """The nametrans-transposed name of the folder's name."""

        return self.visiblename

    def getexplainedname(self):
        """Name that shows both real and nametrans-mangled values."""

        if self.name == self.visiblename:
            return self.name
        else:
            return "%s [remote name %s]"% (self.visiblename, self.name)

    def getrepository(self):
        """Returns the repository object that this folder is within."""

        return self.repository

    def getroot(self):
        """Returns the root of the folder, in a folder-specific fashion."""

        return self.root

    def getsep(self):
        """Returns the separator for this folder type."""

        return self.sep

    def getfullname(self):
        if self.getroot():
            return self.getroot() + self.getsep() + self.getname()
        else:
            return self.getname()

    def getfolderbasename(self):
        """Return base file name of file to store Status/UID info in."""

        if not self.name:
            basename = '.'
        else: # Avoid directory hierarchies and file names such as '/'.
            basename = self.name.replace('/', '.')
        # Replace with literal 'dot' if final path name is '.' as '.' is
        # an invalid file name.
        basename = re.sub('(^|\/)\.$', '\\1dot', basename)
        return basename

    def check_uidvalidity(self):
        """Tests if the cached UIDVALIDITY match the real current one

        If required it saves the UIDVALIDITY value. In this case the
        function is not threadsafe. So don't attempt to call it from
        concurrent threads.

        :returns: Boolean indicating the match. Returns True in case it
            implicitely saved the UIDVALIDITY."""

        if self.get_saveduidvalidity() != None:
            return self.get_saveduidvalidity() == self.get_uidvalidity()
        else:
            self.save_uidvalidity()
            return True

    def _getuidfilename(self):
        """provides UIDVALIDITY cache filename for class internal purposes."""

        return os.path.join(self.repository.getuiddir(),
                            self.getfolderbasename())

    def get_saveduidvalidity(self):
        """Return the previously cached UIDVALIDITY value

        :returns: UIDVALIDITY as (long) number or None, if None had been
            saved yet."""

        if hasattr(self, '_base_saved_uidvalidity'):
            return self._base_saved_uidvalidity
        uidfilename = self._getuidfilename()
        if not os.path.exists(uidfilename):
            self._base_saved_uidvalidity = None
        else:
            file = open(uidfilename, "rt")
            self._base_saved_uidvalidity = int(file.readline().strip())
            file.close()
        return self._base_saved_uidvalidity

    def save_uidvalidity(self):
        """Save the UIDVALIDITY value of the folder to the cache

        This function is not threadsafe, so don't attempt to call it
        from concurrent threads."""

        newval = self.get_uidvalidity()
        uidfilename = self._getuidfilename()

        with open(uidfilename + ".tmp", "wt") as uidfile:
            uidfile.write("%d\n"% newval)
        os.rename(uidfilename + ".tmp", uidfilename)
        self._base_saved_uidvalidity = newval

    def get_uidvalidity(self):
        """Retrieve the current connections UIDVALIDITY value

        This function needs to be implemented by each Backend
        :returns: UIDVALIDITY as a (long) number."""

        raise NotImplementedError

    def cachemessagelist(self):
        """Cache the list of messages.

        Reads the message list from disk or network and stores it in memory for
        later use.  This list will not be re-read from disk or memory unless
        this function is called again."""

        raise NotImplementedError

    def ismessagelistempty(self):
        """Is the list of messages empty."""

        if len(self.messagelist.keys()) < 1:
            return True
        return False

    def dropmessagelistcache(self):
        """Empty everythings we know about messages."""

        self.messagelist = {}

    def getmessagelist(self):
        """Gets the current message list.

        You must call cachemessagelist() before calling this function!"""

        return self.messagelist

    def msglist_item_initializer(self, uid):
        """Returns value for empty messagelist element with given UID.

        This function must initialize all fields of messagelist item
        and must be called every time when one creates new messagelist
        entry to ensure that all fields that must be present are present."""

        raise NotImplementedError

    def uidexists(self, uid):
        """Returns True if uid exists."""

        return uid in self.getmessagelist()

    def getmessageuidlist(self):
        """Gets a list of UIDs.

        You may have to call cachemessagelist() before calling this function!"""

        return sorted(self.getmessagelist().keys())

    def getmessagecount(self):
        """Gets the number of messages."""

        return len(self.getmessagelist())

    def getmessage(self, uid):
        """Returns the content of the specified message."""

        raise NotImplementedError

    def getmaxage(self):
        """Return maxage.

        maxage is allowed to be either an integer or a date of the form
        YYYY-mm-dd. This returns a time_struct."""

        maxagestr = self.config.getdefault("Account %s"%
            self.accountname, "maxage", None)
        if maxagestr is None:
            return None
        # Is it a number?
        try:
            maxage = int(maxagestr)
            if maxage < 1:
                raise OfflineImapError("invalid maxage value %d"% maxage,
                    OfflineImapError.ERROR.MESSAGE)
            return time.gmtime(time.time() - 60*60*24*maxage)
        except ValueError:
            pass # Maybe it was a date.
        # Is it a date string?
        try:
            date = time.strptime(maxagestr, "%Y-%m-%d")
            if date[0] < 1900:
                raise OfflineImapError("maxage led to year %d. "
                    "Abort syncing."% date[0],
                    OfflineImapError.ERROR.MESSAGE)
            if (time.mktime(date) - time.mktime(time.localtime())) > 0:
                raise OfflineImapError("maxage led to future date %s. "
                    "Abort syncing."% maxagestr,
                    OfflineImapError.ERROR.MESSAGE)
            return date
        except ValueError:
            raise OfflineImapError("invalid maxage value %s"% maxagestr,
                OfflineImapError.ERROR.MESSAGE)

    def getmaxsize(self):
        return self.config.getdefaultint("Account %s"%
            self.accountname, "maxsize", None)

    def getstartdate(self):
        """ Retrieve the value of the configuration option startdate """
        datestr = self.config.getdefault("Repository " + self.repository.name,
            'startdate', None)
        try:
            if not datestr:
                return None
            date = time.strptime(datestr, "%Y-%m-%d")
            if date[0] < 1900:
                raise OfflineImapError("startdate led to year %d. "
                    "Abort syncing."% date[0],
                    OfflineImapError.ERROR.MESSAGE)
            if (time.mktime(date) - time.mktime(time.localtime())) > 0:
                raise OfflineImapError("startdate led to future date %s. "
                    "Abort syncing."% datestr,
                    OfflineImapError.ERROR.MESSAGE)
            return date
        except ValueError:
            raise OfflineImapError("invalid startdate value %s",
                OfflineImapError.ERROR.MESSAGE)

    def get_min_uid_file(self):
        startuiddir = os.path.join(self.config.getmetadatadir(),
            'Repository-' + self.repository.name, 'StartUID')
        if not os.path.exists(startuiddir):
            os.mkdir(startuiddir, 0o700)
        return os.path.join(startuiddir, self.getfolderbasename())

    def save_min_uid(self, min_uid):
        uidfile = self.get_min_uid_file()
        fd = open(uidfile, 'wt')
        fd.write(str(min_uid) + "\n")
        fd.close()

    def retrieve_min_uid(self):
        uidfile = self.get_min_uid_file()
        if not os.path.exists(uidfile):
            return None
        try:
            fd = open(uidfile, 'rt')
            min_uid = int(fd.readline().strip())
            fd.close()
            return min_uid
        except:
            raise IOError("Can't read %s"% uidfile)


    def savemessage(self, uid, content, flags, rtime):
        """Writes a new message, with the specified uid.

        If the uid is < 0: The backend should assign a new uid and
           return it.  In case it cannot assign a new uid, it returns
           the negative uid passed in WITHOUT saving the message.

           If the backend CAN assign a new uid, but cannot find out what
           this UID is (as is the case with some IMAP servers), it
           returns 0 but DOES save the message.

           IMAP backend should be the only one that can assign a new
           uid.

        If the uid is > 0, the backend should set the uid to this, if it can.
           If it cannot set the uid to that, it will save it anyway.
           It will return the uid assigned in any case.

        Note that savemessage() does not check against dryrun settings,
        so you need to ensure that savemessage is never called in a
        dryrun mode."""

        raise NotImplementedError

    def getmessagetime(self, uid):
        """Return the received time for the specified message."""

        raise NotImplementedError

    def getmessagemtime(self, uid):
        """Returns the message modification time of the specified message."""

        raise NotImplementedError

    def getmessageflags(self, uid):
        """Returns the flags for the specified message."""

        raise NotImplementedError

    def getmessagekeywords(self, uid):
        """Returns the keywords for the specified message."""

        raise NotImplementedError

    def savemessageflags(self, uid, flags):
        """Sets the specified message's flags to the given set.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        raise NotImplementedError

    def addmessageflags(self, uid, flags):
        """Adds the specified flags to the message's flag set.

        If a given flag is already present, it will not be duplicated.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode.

        :param flags: A set() of flags"""

        newflags = self.getmessageflags(uid) | flags
        self.savemessageflags(uid, newflags)

    def addmessagesflags(self, uidlist, flags):
        """Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        for uid in uidlist:
            if self.uidexists(uid):
                self.addmessageflags(uid, flags)

    def deletemessageflags(self, uid, flags):
        """Removes each flag given from the message's flag set.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode.

        If a given flag is already removed, no action will be taken for that
        flag."""

        newflags = self.getmessageflags(uid) - flags
        self.savemessageflags(uid, newflags)

    def deletemessagesflags(self, uidlist, flags):
        """
        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        for uid in uidlist:
            self.deletemessageflags(uid, flags)

    def getmessagelabels(self, uid):
        """Returns the labels for the specified message."""

        raise NotImplementedError

    def savemessagelabels(self, uid, labels, ignorelabels=set(), mtime=0):
        """Sets the specified message's labels to the given set.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        raise NotImplementedError

    def addmessagelabels(self, uid, labels):
        """Adds the specified labels to the message's labels set.  If a given
        label is already present, it will not be duplicated.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode.

        :param labels: A set() of labels"""

        newlabels = self.getmessagelabels(uid) | labels
        self.savemessagelabels(uid, newlabels)

    def addmessageslabels(self, uidlist, labels):
        """Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        for uid in uidlist:
            self.addmessagelabels(uid, labels)

    def deletemessagelabels(self, uid, labels):
        """Removes each label given from the message's label set.

        If a given label is already removed, no action will be taken for that
        label.

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        newlabels = self.getmessagelabels(uid) - labels
        self.savemessagelabels(uid, newlabels)

    def deletemessageslabels(self, uidlist, labels):
        """
        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        for uid in uidlist:
            self.deletemessagelabels(uid, labels)

    def addmessageheader(self, content, linebreak, headername, headervalue):
        """Adds new header to the provided message.

        WARNING: This function is a bit tricky, and modifying it in the wrong way,
        may easily lead to data-loss.

        Arguments:
        - content: message content, headers and body as a single string
        - linebreak: string that carries line ending
        - headername: name of the header to add
        - headervalue: value of the header to add

        .. note::

           The following documentation will not get displayed correctly after being
           processed by Sphinx. View the source of this method to read it.

        This has to deal with strange corner cases where the header is
        missing or empty.  Here are illustrations for all the cases,
        showing where the header gets inserted and what the end result
        is.  In each illustration, '+' means the added contents.  Note
        that these examples assume LF for linebreak, not CRLF, so '\n'
        denotes a linebreak and '\n\n' corresponds to the transition
        between header and body.  However if the linebreak parameter
        is set to '\r\n' then you would have to substitute '\r\n' for
        '\n' in the below examples.

          * Case 1: No '\n\n', leading '\n'

            +X-Flying-Pig-Header: i am here\n
            \n
            This is the body\n
            next line\n

          * Case 2: '\n\n' at position 0

            +X-Flying-Pig-Header: i am here
            \n
            \n
            This is the body\n
            next line\n

          * Case 3: No '\n\n', no leading '\n'

            +X-Flying-Pig-Header: i am here\n
            +\n
            This is the body\n
            next line\n

          * Case 4: '\n\n' at non-zero position

            Subject: Something wrong with OI\n
            From: some@person.at
            +\nX-Flying-Pig-Header: i am here
            \n
            \n
            This is the body\n
            next line\n
        """

        self.ui.debug('', 'addmessageheader: called to add %s: %s'%
            (headername, headervalue))

        insertionpoint = content.find(linebreak * 2)
        if insertionpoint == -1:
            self.ui.debug('', 'addmessageheader: headers were missing')
        else:
            self.ui.debug('',
                'addmessageheader: headers end at position %d'% insertionpoint)
            mark = '==>EOH<=='
            contextstart = max(0,            insertionpoint - 100)
            contextend   = min(len(content), insertionpoint + 100)
            self.ui.debug('', 'addmessageheader: header/body transition " \
                "context (marked by %s): %s%s%s'% (
                    mark, repr(content[contextstart:insertionpoint]),
                    mark, repr(content[insertionpoint:contextend])
                )
            )

        # Hoping for case #4.
        prefix = linebreak
        suffix = ''
        # Case #2.
        if insertionpoint == 0:
            prefix = ''
            suffix = ''
        # Either case #1 or #3.
        elif insertionpoint == -1:
            prefix = ''
            suffix = linebreak
            insertionpoint = 0
            # Case #3: when body starts immediately, without preceding '\n'
            # (this shouldn't happen with proper mail messages, but
            # we seen many broken ones), we should add '\n' to make
            # new (and the only header, in this case) to be properly
            # separated from the message body.
            if content[0:len(linebreak)] != linebreak:
                suffix = suffix + linebreak

        self.ui.debug('',
            'addmessageheader: insertionpoint = %d'% insertionpoint)
        headers = content[0:insertionpoint]
        self.ui.debug('', 'addmessageheader: headers = %s'% repr(headers))
        new_header = prefix + ("%s: %s"% (headername, headervalue)) + suffix
        self.ui.debug('', 'addmessageheader: new_header = %s'% repr(new_header))
        return headers + new_header + content[insertionpoint:]


    def __find_eoh(self, content):
        """Searches for the point where mail headers end.

        Either double '\n', or end of string.

        Arguments:
        - content: contents of the message to search in
        Returns: position of the first non-header byte.
        """

        eoh_cr = content.find('\n\n')
        if eoh_cr == -1:
            eoh_cr = len(content)

        return eoh_cr


    def getmessageheader(self, content, name):
        """Return the value of the first occurence of the given header.

        Header name is case-insensitive.

        Arguments:
        - contents: message itself
        - name: name of the header to be searched

        Returns: header value or None if no such header was found.
        """

        self.ui.debug('', 'getmessageheader: called to get %s'% name)
        eoh = self.__find_eoh(content)
        self.ui.debug('', 'getmessageheader: eoh = %d'% eoh)
        headers = content[0:eoh]
        self.ui.debug('', 'getmessageheader: headers = %s'% repr(headers))

        m = re.search('^%s:(.*)$'% name, headers,
            flags= re.MULTILINE | re.IGNORECASE)
        if m:
            return m.group(1).strip()
        else:
            return None


    def getmessageheaderlist(self, content, name):
        """Return a list of values for the given header.

        Arguments:
        - contents: message itself
        - name: name of the header to be searched

        Returns: list of header values or empty list if no such header was found.
        """

        self.ui.debug('', 'getmessageheaderlist: called to get %s'% name)
        eoh = self.__find_eoh(content)
        self.ui.debug('', 'getmessageheaderlist: eoh = %d'% eoh)
        headers = content[0:eoh]
        self.ui.debug('', 'getmessageheaderlist: headers = %s'% repr(headers))

        return re.findall('^%s:(.*)$'%
            name, headers, flags= re.MULTILINE | re.IGNORECASE)


    def deletemessageheaders(self, content, header_list):
        """Deletes headers in the given list from the message content.

        Arguments:
        - content: message itself
        - header_list: list of headers to be deleted or just the header name

        We expect our message to have '\n' as line endings."""

        if type(header_list) != type([]):
            header_list = [header_list]
        self.ui.debug('', 'deletemessageheaders: called to delete %s'% (header_list))

        if not len(header_list):
            return content

        eoh = self.__find_eoh(content)
        self.ui.debug('', 'deletemessageheaders: end of headers = %d'% eoh)
        headers = content[0:eoh]
        rest = content[eoh:]
        self.ui.debug('', 'deletemessageheaders: headers = %s'% repr(headers))
        new_headers = []
        for h in headers.split('\n'):
            keep_it = True
            for trim_h in header_list:
                if len(h) > len(trim_h) and h[0:len(trim_h)+1] == (trim_h + ":"):
                    keep_it = False
                    break
            if keep_it:
                new_headers.append(h)

        return '\n'.join(new_headers) + rest




    def change_message_uid(self, uid, new_uid):
        """Change the message from existing uid to new_uid.

        If the backend supports it (IMAP does not).

        :param new_uid: (optional) If given, the old UID will be changed
            to a new UID. This allows backends efficient renaming of
            messages if the UID has changed."""

        raise NotImplementedError

    def deletemessage(self, uid):
        """Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        raise NotImplementedError

    def deletemessages(self, uidlist):
        """Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode."""

        for uid in uidlist:
            self.deletemessage(uid)

    def copymessageto(self, uid, dstfolder, statusfolder, register=1):
        """Copies a message from self to dst if needed, updating the status

        Note that this function does not check against dryrun settings,
        so you need to ensure that it is never called in a
        dryrun mode.

        :param uid: uid of the message to be copied.
        :param dstfolder: A BaseFolder-derived instance
        :param statusfolder: A LocalStatusFolder instance
        :param register: whether we should register a new thread."
        :returns: Nothing on success, or raises an Exception."""

        # Sometimes, it could be the case that if a sync takes awhile,
        # a message might be deleted from the maildir before it can be
        # synced to the status cache.  This is only a problem with
        # self.getmessage().  So, don't call self.getmessage unless
        # really needed.
        if register: # Output that we start a new thread.
            self.ui.registerthread(self.repository.account)

        try:
            message = None
            flags = self.getmessageflags(uid)
            rtime = self.getmessagetime(uid)

            # If any of the destinations actually stores the message body,
            # load it up.
            if dstfolder.storesmessages():
                message = self.getmessage(uid)
            # Succeeded? -> IMAP actually assigned a UID. If newid
            # remained negative, no server was willing to assign us an
            # UID. If newid is 0, saving succeeded, but we could not
            # retrieve the new UID. Ignore message in this case.
            new_uid = dstfolder.savemessage(uid, message, flags, rtime)
            if new_uid > 0:
                if new_uid != uid:
                    # Got new UID, change the local uid to match the new one.
                    self.change_message_uid(uid, new_uid)
                    statusfolder.deletemessage(uid)
                    # Got new UID, change the local uid.
                # Save uploaded status in the statusfolder.
                statusfolder.savemessage(new_uid, message, flags, rtime)
                # Check whether the mail has been seen.
                if 'S' not in flags:
                    self.have_newmail = True
            elif new_uid == 0:
                # Message was stored to dstfolder, but we can't find it's UID
                # This means we can't link current message to the one created
                # in IMAP. So we just delete local message and on next run
                # we'll sync it back
                # XXX This could cause infinite loop on syncing between two
                # IMAP servers ...
                self.deletemessage(uid)
            else:
                raise OfflineImapError("Trying to save msg (uid %d) on folder "
                    "%s returned invalid uid %d"% (uid, dstfolder.getvisiblename(),
                    new_uid), OfflineImapError.ERROR.MESSAGE)
        except (KeyboardInterrupt): # Bubble up CTRL-C.
            raise
        except OfflineImapError as e:
            if e.severity > OfflineImapError.ERROR.MESSAGE:
                raise # Bubble severe errors up.
            self.ui.error(e, exc_info()[2])
        except Exception as e:
            self.ui.error(e, exc_info()[2],
              msg = "Copying message %s [acc: %s]"% (uid, self.accountname))
            raise  # Raise on unknown errors, so we can fix those.

    def __syncmessagesto_copy(self, dstfolder, statusfolder):
        """Pass1: Copy locally existing messages not on the other side.

        This will copy messages to dstfolder that exist locally but are
        not in the statusfolder yet. The strategy is:

        1) Look for messages present in self but not in statusfolder.
        2) invoke copymessageto() on those which:
           - If dstfolder doesn't have it yet, add them to dstfolder.
           - Update statusfolder.

        This function checks and protects us from action in dryrun mode."""

        # We have no new mail yet.
        self.have_newmail = False

        threads = []

        copylist = [uid for uid in self.getmessageuidlist()
            if not statusfolder.uidexists(uid)]
        num_to_copy = len(copylist)

        # Honor 'copy_ignore_eval' configuration option.
        if self.copy_ignoreUIDs is not None:
            for uid in self.copy_ignoreUIDs:
                if uid in copylist:
                    copylist.remove(uid)
                    self.ui.ignorecopyingmessage(uid, self, dstfolder)

        if num_to_copy > 0 and self.repository.account.dryrun:
            self.ui.info("[DRYRUN] Copy {} messages from {}[{}] to {}".format(
                num_to_copy, self, self.repository, dstfolder.repository)
            )
            return

        with self:
            for num, uid in enumerate(copylist):
                # Bail out on CTRL-C or SIGTERM.
                if offlineimap.accounts.Account.abort_NOW_signal.is_set():
                    break

                if uid == 0:
                    self.ui.warn("Assertion that UID != 0 failed; ignoring message.")
                    continue

                if uid > 0 and dstfolder.uidexists(uid):
                    # dstfolder has message with that UID already, only update status.
                    flags = self.getmessageflags(uid)
                    rtime = self.getmessagetime(uid)
                    statusfolder.savemessage(uid, None, flags, rtime)
                    continue

                self.ui.copyingmessage(uid, num+1, num_to_copy, self, dstfolder)
                # Exceptions are caught in copymessageto().
                if self.suggeststhreads():
                    self.waitforthread()
                    thread = threadutil.InstanceLimitedThread(
                        self.getinstancelimitnamespace(),
                        target=self.copymessageto,
                        name="Copy message from %s:%s"% (self.repository, self),
                        args=(uid, dstfolder, statusfolder)
                    )
                    thread.start()
                    threads.append(thread)
                else:
                    self.copymessageto(uid, dstfolder, statusfolder, register=0)
            for thread in threads:
                thread.join() # Block until all "copy" threads are done.

        # Execute new mail hook if we have new mail.
        if self.have_newmail:
            if self.newmail_hook != None:
                self.newmail_hook()

    def __syncmessagesto_delete(self, dstfolder, statusfolder):
        """Pass 2: Remove locally deleted messages on dst.

        Get all UIDs in statusfolder but not self. These are messages
        that were deleted in 'self'. Delete those from dstfolder and
        statusfolder.

        This function checks and protects us from action in dryrun mode.
        """
        # The list of messages to delete. If sync of deletions is disabled we
        # still remove stale entries from statusfolder (neither in local nor
        # remote).
        deletelist = [uid for uid in statusfolder.getmessageuidlist()
                      if uid >= 0 and
                      not self.uidexists(uid) and
                      (self._sync_deletes or not dstfolder.uidexists(uid))]

        if len(deletelist):
            # Delete in statusfolder first to play safe. In case of abort, we
            # won't lose message, we will just unneccessarily retransmit some.
            # Delete messages from statusfolder that were either deleted by the
            # user, or not being tracked (e.g. because of maxage).
            if not self.repository.account.dryrun:
                statusfolder.deletemessages(deletelist)
            # Filter out untracked messages.
            deletelist = [uid for uid in deletelist if dstfolder.uidexists(uid)]
            if len(deletelist):
                self.ui.deletingmessages(deletelist, [dstfolder])
                if not self.repository.account.dryrun:
                    dstfolder.deletemessages(deletelist)

    def combine_flags_and_keywords(self, uid, dstfolder):
        """Combine the message's flags and keywords using the mapping for the
        destination folder."""

        # Take a copy of the message flag set, otherwise
        # __syncmessagesto_flags() will fail because statusflags is actually a
        # reference to selfflags (which it should not, but I don't have time to
        # debug THAT).
        selfflags = set(self.getmessageflags(uid))

        try:
            keywordmap = dstfolder.getrepository().getkeywordmap()
            if keywordmap is None:
                return selfflags

            knownkeywords = set(keywordmap.keys())
            selfkeywords = self.getmessagekeywords(uid)

            if not knownkeywords >= selfkeywords:
                # Some of the message's keywords are not in the mapping, so
                # skip them.
                skipped_keywords = list(selfkeywords - knownkeywords)
                selfkeywords &= knownkeywords
                self.ui.warn("Unknown keywords skipped: %s\n"
                    "You may want to change your configuration to include "
                    "those\n" % (skipped_keywords))

            keywordletterset = set([keywordmap[keyw] for keyw in selfkeywords])

            # Add the mapped keywords to the list of message flags.
            selfflags |= keywordletterset
        except NotImplementedError:
            pass

        return selfflags

    def __syncmessagesto_flags(self, dstfolder, statusfolder):
        """Pass 3: Flag synchronization.

        Compare flag mismatches in self with those in statusfolder. If
        msg has a valid UID and exists on dstfolder (has not e.g. been
        deleted there), sync the flag change to both dstfolder and
        statusfolder.

        This function checks and protects us from action in ryrun mode.
        """

        # For each flag, we store a list of uids to which it should be
        # added.  Then, we can call addmessagesflags() to apply them in
        # bulk, rather than one call per message.
        addflaglist = {}
        delflaglist = {}
        for uid in self.getmessageuidlist():
            # Ignore messages with negative UIDs missed by pass 1 and
            # don't do anything if the message has been deleted remotely
            if uid < 0 or not dstfolder.uidexists(uid):
                continue

            if statusfolder.uidexists(uid):
                statusflags = statusfolder.getmessageflags(uid)
            else:
                statusflags = set()

            selfflags = self.combine_flags_and_keywords(uid, dstfolder)

            addflags = selfflags - statusflags
            delflags = statusflags - selfflags

            for flag in addflags:
                if not flag in addflaglist:
                    addflaglist[flag] = []
                addflaglist[flag].append(uid)

            for flag in delflags:
                if not flag in delflaglist:
                    delflaglist[flag] = []
                delflaglist[flag].append(uid)

        for flag, uids in addflaglist.items():
            self.ui.addingflags(uids, flag, dstfolder)
            if self.repository.account.dryrun:
                continue # Don't actually add in a dryrun.
            dstfolder.addmessagesflags(uids, set(flag))
            statusfolder.addmessagesflags(uids, set(flag))

        for flag, uids in delflaglist.items():
            self.ui.deletingflags(uids, flag, dstfolder)
            if self.repository.account.dryrun:
                continue # Don't actually remove in a dryrun.
            dstfolder.deletemessagesflags(uids, set(flag))
            statusfolder.deletemessagesflags(uids, set(flag))

    def syncmessagesto(self, dstfolder, statusfolder):
        """Syncs messages in this folder to the destination dstfolder.

        This is the high level entry for syncing messages in one direction.
        Syncsteps are:

        Pass1: Copy locally existing messages
         Copy messages in self, but not statusfolder to dstfolder if not
         already in dstfolder. dstfolder might assign a new UID (e.g. if
         uploading to IMAP). Update statusfolder.

        Pass2: Remove locally deleted messages
         Get all UIDS in statusfolder but not self. These are messages
         that were deleted in 'self'. Delete those from dstfolder and
         statusfolder.

         After this pass, the message lists should be identical wrt the
         uids present (except for potential negative uids that couldn't
         be placed anywhere).

        Pass3: Synchronize flag changes
         Compare flag mismatches in self with those in statusfolder. If
         msg has a valid UID and exists on dstfolder (has not e.g. been
         deleted there), sync the flag change to both dstfolder and
         statusfolder.

        Pass4: Synchronize label changes (Gmail only)
         Compares label mismatches in self with those in statusfolder.
         If msg has a valid UID and exists on dstfolder, syncs the labels
         to both dstfolder and statusfolder.

        :param dstfolder: Folderinstance to sync the msgs to.
        :param statusfolder: LocalStatus instance to sync against.
        """

        for action in self.syncmessagesto_passes:
            # Bail out on CTRL-C or SIGTERM.
            if offlineimap.accounts.Account.abort_NOW_signal.is_set():
                break
            try:
                action(dstfolder, statusfolder)
            except (KeyboardInterrupt):
                raise
            except OfflineImapError as e:
                if e.severity > OfflineImapError.ERROR.FOLDER:
                    raise
                self.ui.error(e, exc_info()[2], "while syncing %s [account %s]"%
                                                (self, self.accountname))
            except Exception as e:
                self.ui.error(e, exc_info()[2], "while syncing %s [account %s]"%
                                                (self, self.accountname))
                raise # Raise unknown Exceptions so we can fix them.

    def __eq__(self, other):
        """Comparisons work either on string comparing folder names or
        on the same instance.

        MailDirFolder('foo') == 'foo' --> True
        a = MailDirFolder('foo'); a == b --> True
        MailDirFolder('foo') == 'moo' --> False
        MailDirFolder('foo') == IMAPFolder('foo') --> False
        MailDirFolder('foo') == MaildirFolder('foo') --> False
        """

        if isinstance(other, str):
            return other == self.name
        return id(self) == id(other)

    def __ne__(self, other):
        return not self.__eq__(other)



# Some useful functions to extract data out of emails
# Copyright (C) 2002-2015 John Goerzen & contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import email
from email.parser import Parser as MailParser

def get_message_date(content, header='Date'):
    """Parses mail and returns resulting timestamp.

    :param header: the header to extract date from;
    :returns: timestamp or `None` in the case of failure.
    """

    message = MailParser().parsestr(content, True)
    dateheader = message.get(header)
    # parsedate_tz returns a 10-tuple that can be passed to mktime_tz
    # Will be None if missing or not in a valid format.  Note that
    # indexes 6, 7, and 8 of the result tuple are not usable.
    datetuple = email.utils.parsedate_tz(dateheader)
    if datetuple is None:
        return None
    return email.utils.mktime_tz(datetuple)



# Copyright (C) 2002-2016 John Goerzen & contributors.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import six
from sys import exc_info

try:
    from configparser import NoSectionError
except ImportError: #python2
    from ConfigParser import NoSectionError

from offlineimap.repository.IMAP import IMAPRepository, MappedIMAPRepository
from offlineimap.repository.Gmail import GmailRepository
from offlineimap.repository.Maildir import MaildirRepository
from offlineimap.repository.GmailMaildir import GmailMaildirRepository
from offlineimap.repository.LocalStatus import LocalStatusRepository
from offlineimap.error import OfflineImapError


class Repository(object):
    """Abstract class that returns the correct Repository type
    instance based on 'account' and 'reqtype', e.g.  a
    class:`ImapRepository` instance."""

    def __new__(cls, account, reqtype):
        """
        :param account: :class:`Account`
        :param regtype: 'remote', 'local', or 'status'"""

        if reqtype == 'remote':
            name = account.getconf('remoterepository')
            # We don't support Maildirs on the remote side.
            typemap = {'IMAP': IMAPRepository,
                'Gmail': GmailRepository}

        elif reqtype == 'local':
            name = account.getconf('localrepository')
            typemap = {'IMAP': MappedIMAPRepository,
                'Maildir': MaildirRepository,
                'GmailMaildir': GmailMaildirRepository}

        elif reqtype == 'status':
            # create and return a LocalStatusRepository.
            name = account.getconf('localrepository')
            return LocalStatusRepository(name, account)

        else:
            errstr = "Repository type %s not supported" % reqtype
            raise OfflineImapError(errstr, OfflineImapError.ERROR.REPO)

        # Get repository type.
        config = account.getconfig()
        try:
            repostype = config.get('Repository ' + name, 'type').strip()
        except NoSectionError as e:
            errstr = ("Could not find section '%s' in configuration. Required "
                      "for account '%s'." % ('Repository %s' % name, account))
            six.reraise(OfflineImapError,
                        OfflineImapError(errstr, OfflineImapError.ERROR.REPO),
                        exc_info()[2])

        try:
            repo = typemap[repostype]
        except KeyError:
            errstr = "'%s' repository not supported for '%s' repositories."% \
                (repostype, reqtype)
            six.reraise(OfflineImapError,
                        OfflineImapError(errstr, OfflineImapError.ERROR.REPO),
                        exc_info()[2])

        return repo(name, account)

    def __init__(self, account, reqtype):
        """Load the correct Repository type and return that. The
        __init__ of the corresponding Repository class will be
        executed instead of this stub

        :param account: :class:`Account`
        :param regtype: 'remote', 'local', or 'status'
        """
        pass



# Copyright 2006-2018 Eygene A. Ryabinkin & contributors.
#
# Module that supports distribution-specific functions.

import platform
import os


# Each dictionary value is either string or some iterable.
#
# For the former we will just return the value, for an iterable
# we will walk through the values and will return the first
# one that corresponds to the existing file.
__DEF_OS_LOCATIONS = {
    'freebsd': '/usr/local/share/certs/ca-root-nss.crt',
    'openbsd': '/etc/ssl/cert.pem',
    'dragonfly': '/etc/ssl/cert.pem',
    'darwin': [
      # MacPorts, port curl-ca-bundle
      '/opt/local/share/curl/curl-ca-bundle.crt',
      # homebrew, package openssl
      '/usr/local/etc/openssl/cert.pem',
    ],
    'linux-ubuntu': '/etc/ssl/certs/ca-certificates.crt',
    'linux-debian': '/etc/ssl/certs/ca-certificates.crt',
    'linux-gentoo': '/etc/ssl/certs/ca-certificates.crt',
    'linux-fedora': '/etc/pki/tls/certs/ca-bundle.crt',
    'linux-redhat': '/etc/pki/tls/certs/ca-bundle.crt',
    'linux-suse': '/etc/ssl/ca-bundle.pem',
    'linux-opensuse': '/etc/ssl/ca-bundle.pem',
    'linux-arch': '/etc/ssl/certs/ca-certificates.crt',
}


def get_os_name():
    """
    Finds out OS name.  For non-Linux system it will be just a plain
    OS name (like FreeBSD), for Linux it will be "linux-<distro>",
    where <distro> is the name of the distribution, as returned by
    the first component of platform.linux_distribution.

    Return value will be all-lowercase to avoid confusion about
    proper name capitalisation.

    """
    OS = platform.system().lower()

    if OS.startswith('linux'):
        DISTRO = platform.linux_distribution()[0]
        if DISTRO:
            OS = OS + "-%s" % DISTRO.split()[0].lower()
        if os.path.exists('/etc/arch-release'):
            OS = "linux-arch"

    return OS

def get_os_sslcertfile_searchpath():
    """Returns search path for CA bundle for the current OS.

    We will return an iterable even if configuration has just
    a single value: it is easier for our callers to be sure
    that they can iterate over result.

    Returned value of None means that there is no search path
    at all.
    """

    OS = get_os_name()

    l = None
    if OS in __DEF_OS_LOCATIONS:
        l = __DEF_OS_LOCATIONS[OS]
        if not hasattr(l, '__iter__'):
            l = (l, )
    return l


def get_os_sslcertfile():
    """
    Finds out the location for the distribution-specific
    CA certificate file bundle.

    Returns the location of the file or None if there is
    no known CA certificate file or all known locations
    correspond to non-existing filesystem objects.
    """

    l = get_os_sslcertfile_searchpath()
    if l == None:
        return None

    for f in l:
      assert (type(f) == type(""))
      if os.path.exists(f) and \
        (os.path.isfile(f) or os.path.islink(f)):
          return f

    return None



# Copyright (C) 2013-2014 Eygene A. Ryabinkin and contributors
#
# Collection of classes that implement const-like behaviour
# for various objects.

import copy

class ConstProxy(object):
    """Implements read-only access to a given object
    that can be attached to each instance only once."""

    def __init__(self):
        self.__dict__['__source'] = None


    def __getattr__(self, name):
        src = self.__dict__['__source']
        if src == None:
            raise ValueError("using non-initialized ConstProxy() object")
        return copy.deepcopy(getattr(src, name))


    def __setattr__(self, name, value):
        raise AttributeError("tried to set '%s' to '%s' for constant object"% \
            (name, value))


    def __delattr__(self, name):
        raise RuntimeError("tried to delete field '%s' from constant object"% \
            (name))


    def set_source(self, source):
        """ Sets source object for this instance. """
        if (self.__dict__['__source'] != None):
            raise ValueError("source object is already set")
        self.__dict__['__source'] = source



# Copyright (C) 2007-2018 John Goerzen & contributors.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
try:
    from urllib import urlencode
except ImportError: # python3
    from urllib.parse import urlencode
import sys
import time
import logging
from threading import currentThread

import offlineimap
from offlineimap.ui.UIBase import UIBase

protocol = '7.2.0'

class MachineLogFormatter(logging.Formatter):
    """urlencodes any outputted line, to avoid multi-line output"""

    def format(s, record):
        # Mapping of log levels to historic tag names
        severity_map = {
            'info': 'msg',
            'warning': 'warn',
        }
        line = super(MachineLogFormatter, s).format(record)
        severity = record.levelname.lower()
        if severity in severity_map:
            severity = severity_map[severity]
        if hasattr(record, "machineui"):
            command = record.machineui["command"]
            whoami = record.machineui["id"]
        else:
            command = ""
            whoami = currentThread().getName()

        prefix = "%s:%s"% (command, urlencode([('', whoami)])[1:])
        return "%s:%s:%s"% (severity, prefix, urlencode([('', line)])[1:])


class MachineUI(UIBase):
    def __init__(s, config, loglevel=logging.INFO):
        super(MachineUI, s).__init__(config, loglevel)
        s._log_con_handler.createLock()
        """lock needed to block on password input"""
        # Set up the formatter that urlencodes the strings...
        s._log_con_handler.setFormatter(MachineLogFormatter())

    # Arguments:
    # - handler: must be method from s.logger that reflects
    #   the severity of the passed message
    # - command: command that produced this message
    # - msg: the message itself
    def _printData(s, handler, command, msg):
        handler(msg,
                extra = {
                  'machineui': {
                   'command': command,
                   'id': currentThread().getName(),
                  }
                })

    def _msg(s, msg):
        s._printData(s.logger.info, '_display', msg)

    def warn(s, msg, minor=0):
        # TODO, remove and cleanup the unused minor stuff
        s._printData(s.logger.warning, '', msg)

    def registerthread(s, account):
        super(MachineUI, s).registerthread(account)
        s._printData(s.logger.info, 'registerthread', account)

    def unregisterthread(s, thread):
        UIBase.unregisterthread(s, thread)
        s._printData(s.logger.info, 'unregisterthread', thread.getName())

    def debugging(s, debugtype):
        s._printData(s.logger.debug, 'debugging', debugtype)

    def acct(s, accountname):
        s._printData(s.logger.info, 'acct', accountname)

    def acctdone(s, accountname):
        s._printData(s.logger.info, 'acctdone', accountname)

    def validityproblem(s, folder):
        s._printData(s.logger.warning, 'validityproblem', "%s\n%s\n%s\n%s"%
                (folder.getname(), folder.getrepository().getname(),
                 folder.get_saveduidvalidity(), folder.get_uidvalidity()))

    def connecting(s, reposname, hostname, port):
        s._printData(s.logger.info, 'connecting', "%s\n%s\n%s"% (hostname,
            str(port), reposname))

    def syncfolders(s, srcrepos, destrepos):
        s._printData(s.logger.info, 'syncfolders', "%s\n%s"% (s.getnicename(srcrepos),
                                                s.getnicename(destrepos)))

    def syncingfolder(s, srcrepos, srcfolder, destrepos, destfolder):
        s._printData(s.logger.info, 'syncingfolder', "%s\n%s\n%s\n%s\n"%
                (s.getnicename(srcrepos), srcfolder.getname(),
                 s.getnicename(destrepos), destfolder.getname()))

    def loadmessagelist(s, repos, folder):
        s._printData(s.logger.info, 'loadmessagelist', "%s\n%s"% (s.getnicename(repos),
                                                    folder.getvisiblename()))

    def messagelistloaded(s, repos, folder, count):
        s._printData(s.logger.info, 'messagelistloaded', "%s\n%s\n%d"%
                (s.getnicename(repos), folder.getname(), count))

    def syncingmessages(s, sr, sf, dr, df):
        s._printData(s.logger.info, 'syncingmessages', "%s\n%s\n%s\n%s\n"%
                (s.getnicename(sr), sf.getname(), s.getnicename(dr),
                 df.getname()))

    def ignorecopyingmessage(s, uid, srcfolder, destfolder):
        s._printData(s.logger.info, 'ignorecopyingmessage', "%d\n%s\n%s\n%s[%s]"%
                (uid, s.getnicename(srcfolder), srcfolder.getname(),
                 s.getnicename(destfolder), destfolder))

    def copyingmessage(s, uid, num, num_to_copy, srcfolder, destfolder):
        s._printData(s.logger.info, 'copyingmessage', "%d\n%s\n%s\n%s[%s]"%
                (uid, s.getnicename(srcfolder), srcfolder.getname(),
                 s.getnicename(destfolder), destfolder))

    def folderlist(s, list):
        return ("\f".join(["%s\t%s"% (s.getnicename(x), x.getname()) for x in list]))

    def uidlist(s, list):
        return ("\f".join([str(u) for u in list]))

    def deletingmessages(s, uidlist, destlist):
        ds = s.folderlist(destlist)
        s._printData(s.logger.info, 'deletingmessages', "%s\n%s"% (s.uidlist(uidlist), ds))

    def addingflags(s, uidlist, flags, dest):
        s._printData(s.logger.info, "addingflags", "%s\n%s\n%s"% (s.uidlist(uidlist),
                                                    "\f".join(flags),
                                                    dest))

    def deletingflags(s, uidlist, flags, dest):
        s._printData(s.logger.info, 'deletingflags', "%s\n%s\n%s"% (s.uidlist(uidlist),
                                                      "\f".join(flags),
                                                      dest))

    def threadException(s, thread):
        s._printData(s.logger.warning, 'threadException', "%s\n%s"%
                     (thread.getName(), s.getThreadExceptionString(thread)))
        s.delThreadDebugLog(thread)
        s.terminate(100)

    def terminate(s, exitstatus=0, errortitle='', errormsg=''):
        s._printData(s.logger.info, 'terminate', "%d\n%s\n%s"% (exitstatus, errortitle, errormsg))
        sys.exit(exitstatus)

    def mainException(s):
        s._printData(s.logger.warning, 'mainException', s.getMainExceptionString())

    def threadExited(s, thread):
        s._printData(s.logger.info, 'threadExited', thread.getName())
        UIBase.threadExited(s, thread)

    def sleeping(s, sleepsecs, remainingsecs):
        s._printData(s.logger.info, 'sleeping', "%d\n%d"% (sleepsecs, remainingsecs))
        if sleepsecs > 0:
            time.sleep(sleepsecs)
        return 0


    def getpass(s, username, config, errmsg=None):
        if errmsg:
            s._printData(s.logger.warning,
              'getpasserror', "%s\n%s"% (username, errmsg),
              False)

        s._log_con_handler.acquire() # lock the console output
        try:
            s._printData(s.logger.info, 'getpass', username)
            return (sys.stdin.readline()[:-1])
        finally:
            s._log_con_handler.release()

    def init_banner(s):
        s._printData(s.logger.info, 'protocol', protocol)
        s._printData(s.logger.info, 'initbanner', offlineimap.banner)

    def callhook(s, msg):
        s._printData(s.logger.info, 'callhook', msg)



# Gmail IMAP repository support
# Copyright (C) 2008-2016 Riccardo Murri <riccardo.murri@gmail.com> &
# contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

from offlineimap.repository.IMAP import IMAPRepository
from offlineimap import folder, OfflineImapError


class GmailRepository(IMAPRepository):
    """Gmail IMAP repository.

    This class just has default settings for GMail's IMAP service. So
    you can do 'type = Gmail' instead of 'type = IMAP' and skip
    specifying the hostname, port etc. See
    http://mail.google.com/support/bin/answer.py?answer=78799&topic=12814
    for the values we use."""
    def __init__(self, reposname, account):
        """Initialize a GmailRepository object."""
        IMAPRepository.__init__(self, reposname, account)

    def gethost(self):
        """Return the server name to connect to.

        We first check the usual IMAP settings, and then fall back to
        imap.gmail.com if nothing is specified."""
        try:
            return super(GmailRepository, self).gethost()
        except OfflineImapError:
            # Nothing was configured, cache and return hardcoded
            # one. See the parent class (IMAPRepository) for how this
            # cache is used.
            self._host = "imap.gmail.com"
            return self._host

    def getoauth2_request_url(self):
        """Return the OAuth URL to request tokens from.

        We first check the usual OAuth settings, and then fall back to
        https://accounts.google.com/o/oauth2/token if nothing is
        specified."""

        url = super(GmailRepository, self).getoauth2_request_url()
        if url is None:
            # Nothing was configured, cache and return hardcoded one.
            self.setoauth2_request_url("https://accounts.google.com/o/oauth2/token")
        else:
            self.setoauth2_request_url(url)
        return self.oauth2_request_url

    def getport(self):
        """Return the port number to connect to.

        This Gmail implementation first checks for the usual IMAP settings
        and falls back to 993 if nothing is specified."""

        port = super(GmailRepository, self).getport()

        if port is None:
            return 993
        else:
            return port

    def getssl(self):
        ssl = self.getconfboolean('ssl', None)

        if ssl is None:
            # Nothing was configured, return our default setting for
            # GMail. Maybe this should look more similar to gethost &
            # we could just rely on the global "ssl = yes" default.
            return True
        else:
            return ssl

    def getpreauthtunnel(self):
        return None

    def getfolder(self, foldername, decode=True):
        return self.getfoldertype()(self.imapserver, foldername,
                                    self, decode)

    def getfoldertype(self):
        return folder.Gmail.GmailFolder

    def gettrashfolder(self, foldername):
        # Where deleted mail should be moved
        return self.getconf('trashfolder', '[Gmail]/Trash')

    def getspamfolder(self):
        # Depending on the IMAP settings (Settings -> Forwarding and
        # POP/IMAP -> IMAP Access -> "When I mark a message in IMAP as
        # deleted") GMail might also deletes messages upon EXPUNGE in
        # the Spam folder.
        return self.getconf('spamfolder', '[Gmail]/Spam')



# Local status cache repository support
# Copyright (C) 2002-2017 John Goerzen & contributors
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import os

from offlineimap.folder.LocalStatus import LocalStatusFolder
from offlineimap.folder.LocalStatusSQLite import LocalStatusSQLiteFolder
from offlineimap.repository.Base import BaseRepository
from offlineimap.error import OfflineImapError


class LocalStatusRepository(BaseRepository):
    def __init__(self, reposname, account):
        BaseRepository.__init__(self, reposname, account)

        # class and root for all backends.
        self.backends = {}
        self.backends['sqlite'] = {
            'class': LocalStatusSQLiteFolder,
            'root': os.path.join(account.getaccountmeta(), 'LocalStatus-sqlite')
        }
        self.backends['plain'] = {
            'class': LocalStatusFolder,
            'root': os.path.join(account.getaccountmeta(), 'LocalStatus')
        }

        if self.account.getconf('status_backend', None) is not None:
            raise OfflineImapError(
                "the 'status_backend' configuration option is not supported"
                " anymore; please, remove this configuration option.",
                OfflineImapError.ERROR.REPO
            )
        # Set class and root for sqlite.
        self.setup_backend('sqlite')

        if not os.path.exists(self.root):
            os.mkdir(self.root, 0o700)

        # self._folders is a dict of name:LocalStatusFolders().
        self._folders = {}

    def _instanciatefolder(self, foldername):
        return self.LocalStatusFolderClass(foldername, self) # Instanciate.

    def setup_backend(self, backend):
        if backend in self.backends.keys():
            self._backend = backend
            self.root = self.backends[backend]['root']
            self.LocalStatusFolderClass = self.backends[backend]['class']

    def import_other_backend(self, folder):
        for bk, dic in self.backends.items():
            # Skip folder's own type.
            if dic['class'] == type(folder):
                continue

            repobk = LocalStatusRepository(self.name, self.account)
            repobk.setup_backend(bk)      # Fake the backend.
            folderbk = dic['class'](folder.name, repobk)

            # If backend contains data, import it to folder.
            if not folderbk.isnewfolder():
                self.ui._msg("Migrating LocalStatus cache from %s to %s "
                    "status folder for %s:%s"%
                    (bk, self._backend, self.name, folder.name))

                folderbk.cachemessagelist()
                folder.messagelist = folderbk.messagelist
                folder.saveall()
                break

    def getsep(self):
        return '.'

    def makefolder(self, foldername):
        """Create a LocalStatus Folder."""

        if self.account.dryrun:
            return # Bail out in dry-run mode.

        # Create an empty StatusFolder.
        folder = self._instanciatefolder(foldername)
        # First delete any existing data to make sure we won't consider obsolete
        # data. This might happen if the user removed the folder (maildir) and
        # it is re-created afterwards.
        folder.purge()
        folder.openfiles()
        folder.save()
        folder.closefiles()

        # Invalidate the cache.
        self.forgetfolders()

    def getfolder(self, foldername):
        """Return the Folder() object for a foldername.

        Caller must call closefiles() on the folder when done."""

        if foldername in self._folders:
            return self._folders[foldername]

        folder = self._instanciatefolder(foldername)

        # If folder is empty, try to import data from an other backend.
        if folder.isnewfolder():
            self.import_other_backend(folder)

        self._folders[foldername] = folder
        return folder

    def getfolders(self):
        """Returns a list of all cached folders.

        Does nothing for this backend. We mangle the folder file names
        (see getfolderfilename) so we can not derive folder names from
        the file names that we have available. TODO: need to store a
        list of folder names somehow?"""

        pass

    def forgetfolders(self):
        """Forgets the cached list of folders, if any.  Useful to run
        after a sync run."""

        self._folders = {}
